@@ -21,10 +21,15 @@
  */
 
 #include <IOKit/IOLib.h>
+#define IOFRAMEBUFFER_PRIVATE
 #include <IOKit/graphics/IODisplay.h>
 #include <IOKit/ndrvsupport/IOMacOSVideo.h>
 #include <IOKit/pwr_mgt/RootDomain.h>
 
+#ifndef kAppleClamshellStateKey
+#define kAppleClamshellStateKey	"AppleClamshellState"
+#endif
+
 
 /*
     We further divide the actual display panel brightness levels into four
@@ -59,6 +64,10 @@ enum {
     kIOBacklightDisplayMaxUsableState  = kIODisplayMaxPowerState - 1
 };
 
+extern IOOptionBits gIOFBLastClamshellState;
+
+#define kIOBacklightUserBrightnessKey	"IOBacklightUserBrightness"
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 // probe
 //
@@ -110,11 +119,10 @@ IOService * IOBacklightDisplay::probe( IOService * provider, SInt32 * score )
 
 void IOBacklightDisplay::stop( IOService * provider )
 {
-    getPMRootDomain()->receivePowerNotification( kIOPMDisableClamshell );
-
     return( super::stop( provider ));
 }
 
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 // initForPM
 //
@@ -127,13 +135,15 @@ void IOBacklightDisplay::initPowerManagement( IOService * provider )
     // version,
     //   capabilityFlags,	   outputPowerCharacter, inputPowerRequirement,
     { 1, 0,                                     0, 0,           0,0,0,0,0,0,0,0 },
-    { 1, 0,                      		0, IOPMPowerOn, 0,0,0,0,0,0,0,0 },
+    { 1, 0,                      		0, 0,           0,0,0,0,0,0,0,0 },
     { 1, IOPMDeviceUsable,                      0, IOPMPowerOn, 0,0,0,0,0,0,0,0 },
     { 1, IOPMDeviceUsable | IOPMMaxPerformance, 0, IOPMPowerOn, 0,0,0,0,0,0,0,0 }
     // staticPower, unbudgetedPower, powerToAttain, timeToAttain, settleUpTime, 
     // timeToLower, settleDownTime, powerDomainBudget
     };
 
+    OSNumber * num;
+
     if( !fDisplayParams
      || !getIntegerRange( fDisplayParams, gIODisplayBrightnessKey,
                             &fCurrentBrightness, &fMinBrightness, &fMaxBrightness )) {
@@ -146,7 +156,11 @@ void IOBacklightDisplay::initPowerManagement( IOService * provider )
     if( fCurrentBrightness < (fMinBrightness + 2))
         fCurrentBrightness = fMinBrightness + 2;
 
-    fCurrentUserBrightness = fCurrentBrightness;
+    if( (num = OSDynamicCast(OSNumber,
+            getConnection()->getFramebuffer()->getProperty(kIOBacklightUserBrightnessKey))))
+        fCurrentUserBrightness = num->unsigned32BitValue();
+    else
+        fCurrentUserBrightness = fCurrentBrightness;
 
     fMaxBrightnessLevel[0] = 0;
     fMaxBrightnessLevel[1] = fMinBrightness;
@@ -162,8 +176,6 @@ void IOBacklightDisplay::initPowerManagement( IOService * provider )
 
     // register ourselves with policy-maker (us)
     registerPowerDriver(this, (IOPMPowerState *) ourPowerStates, kIODisplayNumPowerStates);
-
-    getPMRootDomain()->receivePowerNotification( kIOPMEnableClamshell );
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -194,6 +206,7 @@ IOReturn IOBacklightDisplay::setPowerState( unsigned long powerState, IOService
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+
 // Alter the backlight brightness by user request.
 
 bool IOBacklightDisplay::doIntegerSet( OSDictionary * params,
@@ -202,13 +215,14 @@ bool IOBacklightDisplay::doIntegerSet( OSDictionary * params,
     if( paramName != gIODisplayBrightnessKey)
         return( super::doIntegerSet( params, paramName, value));
     else {
-        fCurrentUserBrightness = value;
+        if( !gIOFBLastClamshellState)
+            fCurrentUserBrightness = value;
         return( setBrightness( value ));
     }
 }
 
-
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+
 // setBrightness
 
 bool IOBacklightDisplay::setBrightness( SInt32 value )
@@ -314,16 +328,6 @@ unsigned long IOBacklightDisplay::powerStateForDomainState( IOPMPowerFlags domai
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
-class AppleBacklightDisplay : public IOBacklightDisplay
-{
-    OSDeclareDefaultStructors(AppleBacklightDisplay)
-};
-
-#undef super
-#define super IODisplay
-
-OSDefineMetaClassAndStructors(AppleBacklightDisplay, IOBacklightDisplay)
-
 OSMetaClassDefineReservedUnused(IOBacklightDisplay, 0);
 OSMetaClassDefineReservedUnused(IOBacklightDisplay, 1);
 OSMetaClassDefineReservedUnused(IOBacklightDisplay, 2);
@@ -335,3 +339,130 @@ OSMetaClassDefineReservedUnused(IOBacklightDisplay, 7);
 OSMetaClassDefineReservedUnused(IOBacklightDisplay, 8);
 OSMetaClassDefineReservedUnused(IOBacklightDisplay, 9);
 
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+
+class AppleBacklightDisplay : public IOBacklightDisplay
+{
+    OSDeclareDefaultStructors(AppleBacklightDisplay)
+
+protected:
+    IONotifier * fNotify;
+
+public:
+    virtual bool start( IOService * provider );
+    virtual void stop( IOService * provider );
+    // IODisplay
+    virtual void initPowerManagement( IOService * );
+    virtual bool setBrightness( SInt32 value );
+
+private:
+    static bool _clamshellHandler( void * target, void * ref,
+                                    IOService * resourceService );
+};
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+
+#undef super
+#define super IOBacklightDisplay
+
+OSDefineMetaClassAndStructors(AppleBacklightDisplay, IOBacklightDisplay)
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+
+bool AppleBacklightDisplay::start( IOService * provider )
+{
+    if( !super::start( provider))
+        return( false );
+
+    fNotify = addNotification( gIOPublishNotification,
+                        resourceMatching(kAppleClamshellStateKey), 
+                        _clamshellHandler, this, 0, 10000 );
+
+
+    return( true );
+}
+
+void AppleBacklightDisplay::stop( IOService * provider )
+{
+    IOFramebuffer::clamshellEnable( -1 );
+
+    getPMRootDomain()->removeProperty(kAppleClamshellStateKey);
+
+    if( fNotify) {
+        fNotify->remove();
+        fNotify = 0;
+    }
+
+    getConnection()->getFramebuffer()->setProperty(kIOBacklightUserBrightnessKey,
+                                                    fCurrentUserBrightness, 32);
+
+    return( super::stop( provider ));
+}
+
+void AppleBacklightDisplay::initPowerManagement( IOService * provider )
+{
+    super::initPowerManagement( provider );
+
+    IOFramebuffer::clamshellEnable( +1 );
+}
+
+bool AppleBacklightDisplay::_clamshellHandler( void * target, void * ref,
+                                            IOService * resourceService )
+{
+    AppleBacklightDisplay * self = (AppleBacklightDisplay *) target;
+    UInt32		    value;
+    OSObject *		    prop = 0;
+    OSObject *		    clamshellProperty;
+
+
+    clamshellProperty = resourceService->getProperty(kAppleClamshellStateKey);
+    if( !clamshellProperty)
+        return( true );
+
+    value = (clamshellProperty == kOSBooleanTrue);
+    gIOFBLastClamshellState  = value;
+
+//    if( !prop)        return( true );
+
+    getPMRootDomain()->setProperty(kAppleClamshellStateKey, clamshellProperty);
+    resourceService->removeProperty(kAppleClamshellStateKey);
+
+    if( (kOSBooleanTrue == prop) && (kIOPMEnableClamshell & IOFramebuffer::clamshellState())) {
+
+        if( value) {
+#if LCM_HWSLEEP
+            self->fConnection->getFramebuffer()->changePowerStateTo(0);
+#endif
+            self->changePowerStateToPriv( 0 );
+
+        } else {
+            self->changePowerStateToPriv( kIODisplayMaxPowerState );
+#if LCM_HWSLEEP
+            self->fConnection->getFramebuffer()->changePowerStateTo(1);
+#endif
+        }
+    }
+
+    // may be in the right power state already, but wrong brightness because
+    // of the clamshell state at setPowerState time.
+    if( self->fCurrentPowerState) {
+        SInt32 brightness = self->fMaxBrightnessLevel[self->fCurrentPowerState];
+        if( brightness > self->fCurrentUserBrightness)
+            brightness = self->fCurrentUserBrightness;
+        self->setBrightness( brightness );
+    }
+
+    return( true );
+}
+
+bool AppleBacklightDisplay::setBrightness( SInt32 value )
+{
+    if( gIOFBLastClamshellState)
+        value = 0;
+
+#if DEBUG
+    IOLog("brightness[%d,%d] %d\n", fCurrentPowerState, gIOFBLastClamshellState, value);
+#endif
+
+    return( super::setBrightness( value ) );
+}
