@@ -36,6 +36,7 @@
 #include <IOKit/graphics/IOGraphicsTypes.h>
 #include <IOKit/IOPlatformExpert.h>
 #include <IOKit/IODeviceTreeSupport.h>
+#include <IOKit/IOKitKeys.h>
 #include <IOKit/IOLib.h>
 #include <IOKit/assert.h>
 
@@ -66,7 +67,7 @@ const OSSymbol * gIODisplayParametersCommitKey;
 const OSSymbol * gIODisplayParametersDefaultKey;
 
 enum {
-    kIODisplayMaxUsableState  = kIODisplayMaxPowerState
+    kIODisplayMaxUsableState  = kIODisplayMaxPowerState - 1
 };
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -645,8 +646,8 @@ static IOPMPowerState ourPowerStates[kIODisplayNumPowerStates] = {
   // version,
   // capabilityFlags, outputPowerCharacter, inputPowerRequirement,
   { 1, 0,                                     0, 0,           0,0,0,0,0,0,0,0 },
-  { 1, 0,                                     0, IOPMPowerOn, 0,0,0,0,0,0,0,0 },
-  { 1, 0,                                     0, IOPMPowerOn, 0,0,0,0,0,0,0,0 },
+  { 1, 0,                                     0, 0,           0,0,0,0,0,0,0,0 },
+  { 1, IOPMDeviceUsable,                      0, IOPMPowerOn, 0,0,0,0,0,0,0,0 },
   { 1, IOPMDeviceUsable | IOPMMaxPerformance, 0, IOPMPowerOn, 0,0,0,0,0,0,0,0 }
   // staticPower, unbudgetedPower, powerToAttain, timeToAttain, settleUpTime, 
   // timeToLower, settleDownTime, powerDomainBudget
