@@ -141,24 +141,12 @@ IOReturn IOFramebufferUserClient::clientMemoryForType( UInt32 type,
 IOReturn IOFramebufferUserClient::setProperties( OSObject * properties )
 {
     OSDictionary *	props;
-    OSDictionary *	dict;
-    OSArray *		array;
     IOReturn		kr = kIOReturnUnsupported;
 
     if( !(props = OSDynamicCast( OSDictionary, properties)))
 	return( kIOReturnBadArgument);
 
-    if( (dict = OSDynamicCast( OSDictionary, props->getObject(kIOFBConfigKey)))) {
-
-        owner->setProperty( kIOFBConfigKey, dict );
-    
-        if( (array = OSDynamicCast(OSArray,
-                    dict->getObject( kIOFBDetailedTimingsKey))))
-            kr = owner->setDetailedTimings( array );
-        else
-            kr = kIOReturnSuccess;
-
-    }
+    kr = owner->extSetProperties( props );
 
     return( kr );
 }
