@@ -280,7 +280,8 @@ enum {
     kModeSimulscan              = 8,                            /* Indicates that more than one display connection can be driven from a single framebuffer controller. */
     kModeNotPreset              = 9,                            /* Indicates that the timing is not a factory preset for the current display (geometry may need correction) */
     kModeBuiltIn                = 10,                           /* Indicates that the display mode is for the built-in connect only (on multiconnect devices like the PB 3400) Only the driver is quieried */
-    kModeStretched              = 11                            /* Indicates that the display mode will be stretched/distorted to match the display aspect ratio */
+    kModeStretched              = 11,                           /* Indicates that the display mode will be stretched/distorted to match the display aspect ratio */
+    kModeNotGraphicsQuality     = 12                            /* Indicates that the display mode is not the highest quality (eg. stretching artifacts).  Intended as a hint */
 };
 
 /* csDepthFlags in VDVideoParametersInfoRec */
@@ -304,7 +305,18 @@ enum {
     kHardwareWake               = 129,
     kHardwareWakeFromSuspend    = 130,
     kHardwareWakeToDoze         = 131,
-    kHardwareWakeToDozeFromSuspend = 132
+    kHardwareWakeToDozeFromSuspend = 132,
+    kHardwarePark               = 133,
+    kHardwareDrive              = 134
+};
+
+/* Reduced perf level, for GetPowerState, SetPowerState*/
+enum {
+    kPowerStateReducedPowerMask   = 0x00000300,
+    kPowerStateFullPower          = 0x00000000,
+    kPowerStateReducedPower1      = 0x00000100,
+    kPowerStateReducedPower2      = 0x00000200,
+    kPowerStateReducedPower3      = 0x00000300
 };
 
 enum {
@@ -320,7 +332,13 @@ enum {
     kPowerStateSleepForbiddenMask = (1L << kPowerStateSleepForbiddenBit),
     kPowerStateSleepCanPowerOffMask = (1L << kPowerStateSleepCanPowerOffBit),
     kPowerStateSleepNoDPMSMask  = (1L << kPowerStateSleepNoDPMSBit),
-    kPowerStateSleepWaketoDozeMask = (1L << kPowerStateSleepWaketoDozeBit)
+    kPowerStateSleepWaketoDozeMask = (1L << kPowerStateSleepWaketoDozeBit),
+    kPowerStateSupportsReducedPower1Bit = 10,
+    kPowerStateSupportsReducedPower2Bit = 11,
+    kPowerStateSupportsReducedPower3Bit = 12,
+    kPowerStateSupportsReducedPower1BitMask = (1 << 10),
+    kPowerStateSupportsReducedPower2BitMask = (1 << 11),
+    kPowerStateSupportsReducedPower3BitMask = (1 << 12)
 };
 
 
@@ -352,6 +370,8 @@ enum {
     cscProbeConnection		= 34,				/* Takes nil pointer */
                                                                 /* (may generate a kFBConnectInterruptServiceType service interrupt) */
     cscSetScaler		= 36,				/* Takes a VDScalerPtr */
+    cscSetMirror                = 37,				/* Takes a VDMirrorPtr*/
+    cscSetFeatureConfiguration  = 38,				/* Takes a VDConfigurationPtr*/
     cscUnusedCall               = 127                           /* This call used to expand the scrn resource.  Its imbedded data contains more control info */
 };
 
@@ -388,9 +408,11 @@ enum {
     cscGetClutBehavior          = 29,                           /* Takes a VDClutBehavior */
     cscGetTimingRanges          = 30,                           /* Takes a VDDisplayTimingRangePtr */
     cscGetDetailedTiming        = 31,                           /* Takes a VDDetailedTimingPtr */
-    cscGetCommunicationInfo     = 32,                            /* Takes a VDCommunicationInfoPtr */
+    cscGetCommunicationInfo     = 32,                           /* Takes a VDCommunicationInfoPtr */
     cscGetScalerInfo		= 35,				/* Takes a VDScalerInfoPtr */
-    cscGetScaler		= 36				/* Takes a VDScalerPtr */
+    cscGetScaler		= 36,				/* Takes a VDScalerPtr */
+    cscGetMirror                = 37,				/* Takes a VDMirrorPtr*/
+    cscGetFeatureConfiguration  = 38				/* Takes a VDConfigurationPtr*/
 };
 
 /* Bit definitions for the Get/Set Sync call*/
@@ -1063,7 +1085,8 @@ enum {
     kVideoI2CTransactionErr     = -10932,
     kVideoI2CBusyErr            = -10933,
     kVideoI2CTransactionTypeErr = -10934,
-    kVideoBufferSizeErr         = -10935
+    kVideoBufferSizeErr         = -10935,
+    kVideoCannotMirrorErr       = -10936
 };
 
 
@@ -1286,6 +1309,57 @@ struct VDScalerInfoRec {
 typedef struct VDScalerInfoRec   VDScalerInfoRec;
 typedef VDScalerInfoRec *VDScalerInfoPtr;
 
+enum {
+    /* csMirrorFeatures*/
+    kMirrorSameDepthOnlyMirrorMask = (1 << 0),			/* Commonly true - Mirroring can only be done if the displays are the same bitdepth*/
+    kMirrorSameSizeOnlyMirrorMask = (1 << 1),			/* Commonly false - Mirroring can only be done if the displays are the same size*/
+    kMirrorSameTimingOnlyMirrorMask = (1 << 2),			/* Sometimes true - Mirroring can only be done if the displays are the same timing*/
+    kMirrorCommonGammaMask        = (1 << 3)			/* Sometimes true - Only one gamma correction LUT.*/
+};
+
+enum {
+    /* csMirrorSupportedFlags and csMirrorFlags*/
+    kMirrorCanMirrorMask          = (1 << 0), 			/* Set means we can HW mirrored right now (uses csMirrorEntryID)*/
+    kMirrorAreMirroredMask        = (1 << 1), 			/* Set means we are HW mirrored right now (uses csMirrorEntryID)*/
+    kMirrorUnclippedMirrorMask    = (1 << 2), 			/* Set means mirrored displays are not clipped to their intersection*/
+    kMirrorHAlignCenterMirrorMask = (1 << 3), 			/* Set means mirrored displays can/should be centered horizontally*/
+    kMirrorVAlignCenterMirrorMask = (1 << 4), 			/* Set means mirrored displays can/should be centered vertically*/
+    kMirrorCanChangePixelFormatMask = (1 << 5),			/* Set means mirrored the device should change the pixel format of mirrored displays to allow mirroring.*/
+    kMirrorCanChangeTimingMask    = (1 << 6), 			/* Set means mirrored the device should change the timing of mirrored displays to allow mirroring.*/
+    kMirrorClippedMirrorMask	  = (1 << 7)			/* Set means mirrored displays are clipped to their intersection (driver handles blacking and base address adjustment)*/
+};
+
+struct VDMirrorRec {
+    UInt32              csMirrorSize;           /* Init to sizeof(VDMirrorRec)*/
+    UInt32              csMirrorVersion;        /* Init to 0*/
+    
+    void *		csMirrorRequestID[4];   /* Input RegEntryID to check for mirroring support and state*/
+    void *		csMirrorResultID[4];    /* Output RegEntryID of the next mirrored device*/
+    
+    UInt32              csMirrorFeatures;       /* Output summary features of the driver*/
+    UInt32              csMirrorSupportedFlags; /* Output configuration options supported by the driver*/
+    UInt32              csMirrorFlags;          /* Output configuration options active now*/
+    UInt32              csReserved1;            /* Init to 0*/
+    
+    
+    UInt32              csReserved2;            /* Init to 0*/
+    UInt32              csReserved3;            /* Init to 0*/
+    UInt32              csReserved4;            /* Init to 0*/
+    UInt32              csReserved5;            /* Init to 0*/
+};
+typedef struct VDMirrorRec VDMirrorRec;
+typedef VDMirrorRec * VDMirrorPtr;
+
+struct VDConfigurationRec {
+    UInt32              csConfigFeature;        /* input what feature to config - always input*/
+    UInt32              csConfigSupport;        /* output support value - always output*/
+    UInt32              csConfigValue;          /* input/output feature value - input on Control(), output on Status()*/
+    UInt32              csReserved1;
+    UInt32              csReserved2;
+};
+typedef struct VDConfigurationRec       VDConfigurationRec;
+typedef VDConfigurationRec *            VDConfigurationPtr;
+
 #if PRAGMA_STRUCT_ALIGN
     #pragma options align=reset
 #elif PRAGMA_STRUCT_PACKPUSH
